from easy_spark_path.easy_file_type import EasyFileType
from easy_spark_path.easy_warehouse_type import EasyWarehouseType


class EasySparkFilePath:
    LH_DEFAULT_FILE_PATH = '/lakehouse/default/Files'
    WH_DEFAULT_FILE_PATH = '/warehouse/default/Files'

    def __init__(self, path: str, is_relative: bool, extension: str, file_type: EasyFileType,
                 warehouse_type: EasyWarehouseType = EasyWarehouseType.LH):
        self.path = path
        self.extension = extension.upper()
        self.file_type = file_type
        self.is_relative = is_relative
        self.warehouse_type = warehouse_type

    @staticmethod
    def create_from_path(path: str, is_relative: bool, file_type: EasyFileType = None,
                         warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkFilePath':
        extension = EasySparkFilePath.get_extension_from_path(path)
        if extension is None and file_type is not None:
            extension = EasySparkFilePath.get_extension_from_file_type(file_type)
        if extension is None:
            raise ValueError('Extension is required')
        if file_type is None:
            file_type = EasySparkFilePath.get_file_type_from_extension(extension)
        if file_type == EasyFileType.UNKNOWN:
            raise ValueError('Unknown file type')
        new_path = EasySparkFilePath.build_path(path, is_relative, warehouse_type)
        new_path = EasySparkFilePath.build_path_with_extension(new_path, extension)
        return EasySparkFilePath(new_path, is_relative, extension, file_type, warehouse_type)

    @staticmethod
    def create_from_relative_path(path: str, file_type: EasyFileType = None,
                                  warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkFilePath':
        return EasySparkFilePath.create_from_path(path, True, file_type, warehouse_type)

    @staticmethod
    def create_from_absolute_path(path: str, file_type: EasyFileType = None,
                                  warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkFilePath':
        return EasySparkFilePath.create_from_path(path, False, file_type, warehouse_type)

    @staticmethod
    def build_path(path: str, is_relative: bool, warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> str:
        default_path = EasySparkFilePath.LH_DEFAULT_FILE_PATH if warehouse_type == EasyWarehouseType.LH else EasySparkFilePath.WH_DEFAULT_FILE_PATH
        new_path = path

        if is_relative and default_path not in path:
            if new_path[-1] == '/':
                new_path = new_path[:-1]
            if new_path[0] == '/':
                new_path = new_path[1:]
            new_path = f"{default_path}/{path}"

        return new_path

    @staticmethod
    def build_path_with_extension(path: str, extension: str) -> str:
        if '.' in path:
            return path
        return f"{path}.{extension}"

    @staticmethod
    def get_extension_from_path(path: str) -> str | None:
        if '.' not in path:
            return None
        return (path.split('.')[-1]).upper()

    @staticmethod
    def get_file_type_from_extension(extension: str) -> EasyFileType:
        if extension in ['CSV', 'JSON', 'PARQUET', 'AVRO']:
            return EasyFileType(extension)
        return EasyFileType.UNKNOWN

    @staticmethod
    def get_extension_from_file_type(file_type: EasyFileType) -> str:
        return file_type.value.upper()



from easy_sql_data_access.easy_open_data_connection import easy_open_data_connection
from easy_sql_data_access.easy_sql_data_connection_builder import EasySQLDataConnectionBuilder


class EasySQLConnectionFactory:
    def __init__(self, constr: str, autocommit: bool = True):
        self.constr = constr
        self.autocommit = autocommit

    @staticmethod
    def create_from_builder(builder: EasySQLDataConnectionBuilder):
        return EasySQLConnectionFactory(builder.constr)

    def open(self, autocommit: bool = None) -> easy_open_data_connection:
        if autocommit is None:
            autocommit = self.autocommit
        return easy_open_data_connection(self.constr, autocommit)



from easy_spark.easy_spark_instance import EasySparkInstance
from easy_spark_path.easy_spark_table_path import EasySparkTablePath


class EasySparkDFOptimize:
    def __init__(self, spark):
        self.spark_instance = EasySparkInstance(spark)

    def vacuum_from_lh_table_name(self, name: str, table_name: str) -> 'EasySparkDFOptimize':
        vacuum_script = f"VACUUM {name}.{table_name}"
        self.spark_instance.spark.sql(vacuum_script)
        return self

    def vacuum_from_lh_table(self, table_name: str) -> 'EasySparkDFOptimize':
        vacuum_script = f"VACUUM {table_name}"
        self.spark_instance.spark.sql(vacuum_script)
        return self

    def vacuum_from_path(self, path: str) -> 'EasySparkDFOptimize':
        vacuum_script = f"VACUUM delta.`{path}`"
        self.spark_instance.spark.sql(vacuum_script)
        return self

    def vacuum_from_table_pat(self, table_path: EasySparkTablePath) -> 'EasySparkDFOptimize':
        if table_path.is_table_name:
            return self.vacuum_from_lh_table(table_path.path)
        else:
            return self.vacuum_from_path(table_path.path)

    def optimize_from_lh_table_name(self, name: str, table_name: str) -> 'EasySparkDFOptimize':
        optimize_script = f"OPTIMIZE {name}.{table_name}"
        self.spark_instance.spark.sql(optimize_script)
        return self

    def optimize_from_lh_table(self, table_name: str) -> 'EasySparkDFOptimize':
        optimize_script = f"OPTIMIZE {table_name}"
        self.spark_instance.spark.sql(optimize_script)
        return self

    def optimize_from_path(self, path: str) -> 'EasySparkDFOptimize':
        optimize_script = f"OPTIMIZE delta.`{path}`"
        self.spark_instance.spark.sql(optimize_script)
        return self

    def optimize_from_table_pat(self, table_path: EasySparkTablePath) -> 'EasySparkDFOptimize':
        if table_path.is_table_name:
            return self.optimize_from_lh_table(table_path.path)
        else:
            return self.optimize_from_path(table_path.path)



from easy_sql_data_access.easy_open_data_connection import easy_open_data_connection
from easy_sql_data_access.easy_sql_data_access import EasySQLDataAccess
from easy_sql_data_access.easy_sql_data_connection_builder import EasySQLDataConnectionBuilder


class EasySQLManageDataAccess:
    def __init__(self, constr: str, autocommit: bool = True):
        self.constr = constr
        self.autocommit = autocommit

    @staticmethod
    def create_from_builder(builder: EasySQLDataConnectionBuilder, autocommit: bool = True):
        return EasySQLManageDataAccess(builder.constr, autocommit)

    def __enter__(self):
        print('Entering Data Access...')
        self.con_cm = easy_open_data_connection(self.constr, self.autocommit)
        self.con = self.con_cm.__enter__()
        self.data_access = EasySQLDataAccess(self.con)
        return self.data_access

    def __exit__(self, exc_type, exc_val, exc_tb):
        print('Exiting Data Access...')
        self.con_cm.__exit__(exc_type, exc_val, exc_tb)



import requests
import json
from requests import Response

from easy_request.easy_method_type import EasyMethodType


# TODO: Empty result if error
class EasyGetRequest:
    def __init__(self, url: str, method: EasyMethodType = EasyMethodType.GET, headers: dict = None,
                 bearer_token: str = None, root_key: str = None):
        self.headers: dict = {
        }
        self.root_key: str = root_key
        self.body: dict | str = None
        self.url: str = url
        if headers:
            self.headers.update(headers)
        self.method: EasyMethodType = method

        if bearer_token:
            self.headers['Authorization'] = f'Bearer {bearer_token}'

    @staticmethod
    def create(url: str, method: EasyMethodType = EasyMethodType.GET):
        return EasyGetRequest(url, method)

    def set_method(self, method: EasyMethodType) -> 'EasyGetRequest':
        self.method = method
        return self

    def add_headers(self, headers: dict) -> 'EasyGetRequest':
        self.headers.update(headers)
        return self

    def add_header(self, key: str, value: any) -> 'EasyGetRequest':
        self.headers[key] = value
        return self

    def set_bearer_token(self, token: str) -> 'EasyGetRequest':
        self.headers['Authorization'] = f'Bearer {token}'
        return self

    def set_body(self, body: dict | str) -> 'EasyGetRequest':
        self.body = body
        return self

    def set_url(self, url: str) -> 'EasyGetRequest':
        self.url = url
        return self

    def set_root_key(self, root_key: str) -> 'EasyGetRequest':
        self.root_key = root_key
        return self

    def get_request_data(self) -> any:
        response = self.get_request_response()
        response_data = response.json()[self.root_key] if self.root_key else response.json()

        return response_data

    def get_request_response(self) -> Response:
        response: Response = None
        if self.method == EasyMethodType.GET:
            response = self.get_request_with_get_response()
        elif self.method == EasyMethodType.POST:
            response = self.get_request_with_post_response()

        response.raise_for_status()
        if response.status_code == 429:
            raise Exception("Too many requests")

        return response

    def get_request_with_get_response(self) -> Response:
        response = requests.get(self.url, headers=self.headers)

        return response

    def get_request_with_post_response(self) -> Response:
        if isinstance(self.body, str):
            json_body = json.dumps(json.loads(self.body))
        else:
            json_body = json.dumps(self.body)

        request_headers = self.headers
        request_headers['Content-Type'] = 'application/json'

        response = requests.post(self.url, headers=request_headers, data=json_body)

        return response



import inspect


class EasyDataUtils:
    def __init__(self):
        pass

    @staticmethod
    def name_to_pascal_case(name: str) -> str:
        return (name[0].upper() + name[1:]).strip()

    @staticmethod
    def name_to_pascal_case_between_dots(name: str) -> str:
        return '.'.join([EasyDataUtils.name_to_pascal_case(x) for x in name.split('.')])

    @staticmethod
    def names_to_pascal_case_from_dict(record: dict) -> dict:
        return {EasyDataUtils.name_to_pascal_case_between_dots(key): value for key, value in record.items()}

    @staticmethod
    def flatten_record(record: dict):
        out = {}

        def flatten(x, name=''):
            if type(x) is dict:
                for a in x:
                    flatten(x[a], name + a + '.')
            else:
                out[name[:-1]] = x

        flatten(record)

        return

    @staticmethod
    def map_dict_to_object_ctor_simple(cls, record: dict[str, any]):
        return cls(**record)

    @staticmethod
    def map_list_dict_to_object_ctor_simple(cls, records: list[dict[str, any]]):
        return [EasyDataUtils.map_dict_to_object_ctor_simple(cls, record) for record in records]

    @staticmethod
    def map_dict_to_object_ctor(cls, record: dict[str, any]):
        params = inspect.signature(cls.__init__).parameters
        return EasyDataUtils.map_dict_to_object_from_params_ctor(cls, params, record)

    @staticmethod
    def map_dict_to_object_from_params_ctor(cls, params, record: dict[str, any]):
        filtered_data = {key: record[key] for key in params if key in record and key != 'self'}
        return cls(**filtered_data)

    @staticmethod
    def map_object_to_dict(obj: any) -> dict[str, any]:
        return {key: getattr(obj, key) for key in obj.__dict__}

    @staticmethod
    def map_list_dict_to_object_ctor(cls, records: list[dict[str, any]]):
        return [EasyDataUtils.map_dict_to_object_ctor(cls, record) for record in records]

    @staticmethod
    def map_list_dict_to_object_from_params_ctor(cls, params, records: list[dict[str, any]]):
        return [EasyDataUtils.map_dict_to_object_from_params_ctor(cls, params, record) for record in records]

    @staticmethod
    def map_dict_to_object(cls, record: dict[str, any]):
        obj = cls()
        for key, value in record.items():
            setattr(obj, key, value)
        return obj

    @staticmethod
    def map_list_dict_to_object(cls, records: list[dict[str, any]]):
        return [EasyDataUtils.map_dict_to_object(cls, record) for record in records]



from datetime import datetime


class EasyFileDetails:
    def __init__(self, fs_path, file_status, is_directory: bool, file_name: str, full_path: str,
                 readable_modified_time: datetime):
        self.fs_path = fs_path
        self.file_status = file_status
        self.file_name: str = file_name
        self.full_path: str = full_path
        self.readable_modified_time: datetime = readable_modified_time
        self.is_directory: bool = is_directory



from easy_spark_path.easy_spark_file_path import EasySparkFilePath
from easy_spark_path.easy_warehouse_type import EasyWarehouseType


class EasySparkDirectoryPath:
    LH_DEFAULT_FILE_PATH = '/lakehouse/default/Files'
    WH_DEFAULT_FILE_PATH = '/warehouse/default/Files'

    def __init__(self, path: str, is_relative: bool, warehouse_type: EasyWarehouseType = EasyWarehouseType.LH):
        self.path = path
        self.is_relative = is_relative
        self.warehouse_type = warehouse_type

    @staticmethod
    def create_from_path(path: str, is_relative: bool,
                         warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkDirectoryPath':
        new_path = EasySparkFilePath.build_path(path, is_relative, warehouse_type)
        return EasySparkDirectoryPath(new_path, is_relative, warehouse_type)

    @staticmethod
    def create_from_relative_path(path: str,
                                  warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkDirectoryPath':
        return EasySparkDirectoryPath.create_from_path(path, True, warehouse_type)

    @staticmethod
    def create_from_absolute_path(path: str,
                                  warehouse_type: EasyWarehouseType = EasyWarehouseType.LH) -> 'EasySparkDirectoryPath':
        return EasySparkDirectoryPath.create_from_path(path, False, warehouse_type)



# TODO: Add import
from notebookutils import mssparkutils


class EasySparkMSSparkUtilsHelpers:
    def __init__(self):
        pass

    @staticmethod
    def create_directory(path: str):
        mssparkutils.fs.mkdirs(path)



import pandas as pd
from pandas import DataFrame as DataFramePandas
from easy_panda.easy_panda_helpers import EasyPandaHelpers
from easy_panda.easy_panda_type import EasyPandaType
from easy_spark_path.easy_file_type import EasyFileType
from easy_spark_path.easy_spark_file_path import EasySparkFilePath


class EasyPanda:
    def __init__(self, pd_df: DataFramePandas):
        self._pd_df = pd_df

    @staticmethod
    def create(pd_df: DataFramePandas = None) -> 'EasyPanda':
        return EasyPanda(pd_df)

    @property
    def pd_df(self) -> DataFramePandas:
        return self._pd_df

    @pd_df.setter
    def pd_df(self, value: DataFramePandas):
        self._pd_df = value

    def from_dict(self, records: dict) -> 'EasyPanda':
        self._pd_df = pd.DataFrame.from_dict(records)
        return self

    def from_list(self, records: list[dict]) -> 'EasyPanda':
        self._pd_df = pd.DataFrame(records)

        return self

    def from_tuple(self, records: list[tuple]) -> 'EasyPanda':
        self._pd_df = pd.DataFrame(records)

        return self

    def from_json_normalize(self, records: list[dict]) -> 'EasyPanda':
        self._pd_df = pd.json_normalize(records)
        return self

    def from_csv(self, csv_path: str, header=True, seperator=",") -> 'EasyPanda':
        self._pd_df = pd.read_csv(csv_path, header=0 if header else None, sep=seperator)
        return self

    def set_types(self, pd_types: dict[str, EasyPandaType]) -> 'EasyPanda':
        convert_pd_types = {key: pd_types[key].value for key in pd_types}
        convert_pd_types = EasyPandaHelpers.order_pd_types_to_pd_df(self._pd_df, convert_pd_types)
        self._pd_df = self._pd_df.astype(convert_pd_types)
        return self

    def overwrite_types(self, pd_types: dict[str, EasyPandaType]) -> 'EasyPanda':
        convert_pd_types = {key: pd_types[key].value for key in pd_types}
        convert_pd_types = EasyPandaHelpers.overwrite_pd_types(self._pd_df.dtypes.to_dict(), convert_pd_types)
        # convert_pd_types = EasyPandaHelpers.order_pd_types_to_pd_df(self._pd_df, convert_pd_types)
        self._pd_df = self._pd_df.astype(convert_pd_types)
        return self

    def to_dict(self) -> dict:
        return self._pd_df.to_dict()

    def to_list_dict(self) -> list[dict]:
        return self._pd_df.to_dict(orient='records')

    def to_csv(self, path: str, index=False) -> 'EasyPanda':
        self._pd_df.to_csv(path, index=index)
        return self

    def to_json(self, path: str, index=False) -> 'EasyPanda':
        self._pd_df.to_json(path, index=index)
        return self

    def to_parquet(self, path: str, index=False) -> 'EasyPanda':
        self._pd_df.to_parquet(path, index=index)
        return self

    def to_file_path(self, file_path: EasySparkFilePath, index=False) -> 'EasyPanda':
        if file_path.file_type == EasyFileType.CSV:
            return self.to_csv(file_path.path, index)
        elif file_path.file_type == EasyFileType.JSON:
            return self.to_json(file_path.path, index)
        elif file_path.file_type == EasyFileType.PARQUET:
            return self.to_parquet(file_path.path, index)
        else:
            raise ValueError('File type not supported')

    def overwrite_types_from_type(self, from_type: EasyPandaType, to_type: EasyPandaType):
        cols = [col for col, col_type in self._pd_df.dtypes.items() if col_type == from_type.value]
        self._pd_df[cols] = self._pd_df[cols].astype(to_type.value)

    def is_empty_or_null(self) -> bool:
        return self._pd_df.empty or self._pd_df.isnull().values.any()



from easy_sql_data_access.easy_sql_data_connection_builder import EasySQLDataConnectionBuilder
from easy_sql_data_access.easy_sql_manage_data_access import EasySQLManageDataAccess


class EasySQLDataAccessFactory:
    def __init__(self, constr: str, autocommit: bool = True):
        self.constr = constr
        self.autocommit = autocommit

    @staticmethod
    def create_from_builder(builder: EasySQLDataConnectionBuilder):
        return EasySQLDataAccessFactory(builder.constr)

    def open(self, autocommit: bool = None) -> EasySQLManageDataAccess:
        if autocommit is None:
            autocommit = self.autocommit
        return EasySQLManageDataAccess(self.constr, autocommit)



from easy_spark.easy_con_opetator import EasyConOperator


class EasyConValue:
    def __init__(self, value: any, operator: EasyConOperator):
        self.value = value
        self.operator = operator



from pyspark.sql import SparkSession

from easy_spark.easy_merge_option_type import EasyMergeOptionType
from easy_spark.easy_merge_type import EasyMergeType
from easy_spark.easy_save_mode_type import EasySaveModeType
from easy_spark.easy_spark_delta_helpers import EasySparkDeltaHelpers
from easy_spark.easy_spark_df import EasySparkDF
from easy_spark.easy_spark_df_optimize import EasySparkDFOptimize
from easy_spark.easy_spark_instance import EasySparkInstance
from pyspark.sql.functions import *
from delta.tables import *
from pyspark.sql import Row
from pandas import DataFrame as DataFramePandas
import pandas as pd
from easy_spark_fs.easy_spark_catalog import EasySparkCatalog
from easy_spark_fs.easy_spark_fs_file import EasySparkFSFile
from easy_spark_fs.easy_spark_fs_path import EasySparkFSPath
from easy_spark_path.easy_spark_table_path import EasySparkTablePath


# TODO: Use with
# TODO: Add Z Index
# TODO Add Method to remove nulls from df
# TODO: Also support group of operators
# TODO: More SQL queries like delete, join
# TODO: Do the insert
# TODO: Check all the imports
# TODO: For merge you can specify what records you want to insert
# TODO: Overwrite eq and + -
# TODO:
# filtered_df = easy_df.df.filter(
#             (col('LastImpression').isNull()) |  # Filters None values
#             (col('LastImpression') == '') |  # Filters empty strings
#             (isnan(col('LastImpression')))  # Filters NaN valuesA
#         )
##TODO: .option('optimize',True).option('zorderCol',UniqueIdentifierColumn)
# TODO: Joins in DFS method
# TODO: Add methods for alias (df_alias = df.select(col("name").alias("person_name")))
# TODO: Add Warehouse to DF
# TODO: This how you remove duplicates
# ## Remove duplicates based unique identifier
# # Define window specification based on partition column and unique identifier column.
# if SourceDateTimeColumn != 'NA':
#     window_spec = Window.partitionBy(UniqueIdentifierColumn).orderBy(col(SourceDateTimeColumn).desc())
#     # Add a new column "row_num" to identify duplicate rows within each partition.
#     df = df.withColumn("row_num", row_number().over(window_spec))
#     # Keep only rows where "row_num" is 1 (i.e., the row with max value in PartitionColumn).
#     df = df.filter(df.row_num == 1)
#     # Drop the "row_num" column since we no longer need it.
#     df = df.drop("row_num")


class EasySparkDeltaTable:
    _spark: SparkSession = None

    def __init__(self, spark: SparkSession, path: str = None, for_name: bool = False,
                 create_schema: StructType = None,
                 partition_columns: list[str] = None):
        self.spark_instance = EasySparkInstance(spark)
        self.spark_fs_path = EasySparkFSPath(spark)
        self.spark_catalog = EasySparkCatalog(spark)
        self.df_optimize = EasySparkDFOptimize(spark)
        EasySparkDeltaTable._spark = spark
        self._delta_path = None
        self.path: str = path
        self.for_name: bool = for_name

        if self.path:
            if for_name:
                self.for_name_from_path(self.path)
            else:
                self.from_path(self.path)

            if create_schema is not None:
                self.create_empty_if_not_exists(create_schema, partition_columns=partition_columns)
            self.init_delta()

    @staticmethod
    def create(spark: SparkSession, path: str = None, create_schema: StructType = None,
               partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        return EasySparkDeltaTable(spark, path, False, create_schema, partition_columns)

    @staticmethod
    def create_for_name(spark: SparkSession, path: str = None, create_schema: StructType = None,
                        partition_columns: list[str] = None) -> 'EasySparkDeltaTable':

        return EasySparkDeltaTable(spark, path, True, create_schema, partition_columns)

    @staticmethod
    def create_for_name_from_lh_table_name(spark: SparkSession, name: str, table_name: str,
                                           create_schema: StructType = None,
                                           partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        path = f'{name}.{table_name}'
        return EasySparkDeltaTable.create_for_name(spark, path, create_schema, partition_columns)

    @staticmethod
    def create_for_name_from_lh_table(spark: SparkSession, table_name: str,
                                      create_schema: StructType = None,
                                      partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        return EasySparkDeltaTable.create_for_name(spark, table_name, create_schema, partition_columns)

    @staticmethod
    def create_from_table_path(spark: SparkSession, table_path: EasySparkTablePath, create_schema: StructType = None,
                               partition_columns: list[str] = None):
        if table_path.is_table_name:
            return EasySparkDeltaTable.create_for_name_from_lh_table(spark, table_path.path, create_schema,
                                                                     partition_columns)
        else:
            return EasySparkDeltaTable.create(spark, table_path.path, create_schema, partition_columns)

    @property
    def delta_path(self):
        return self._delta_path

    def to_easy_df(self) -> EasySparkDF:
        return EasySparkDF(self._delta_path.toDF(), EasySparkDeltaTable._spark)

    def to_fs_file(self, must_java_import=False) -> EasySparkFSFile:
        return EasySparkFSFile(EasySparkDeltaTable._spark, self.path, must_java_import)

    def from_path(self, path: str) -> 'EasySparkDeltaTable':
        self.path = path

        return self

    def for_name_from_path(self, path: str) -> 'EasySparkDeltaTable':
        self.path = path
        self.for_name = True

        return self

    def init_delta(self) -> 'EasySparkDeltaTable':
        if self.exists():
            if self.for_name:
                self._delta_path = DeltaTable.forName(EasySparkDeltaTable._spark, self.path)
            else:
                self._delta_path = DeltaTable.forPath(EasySparkDeltaTable._spark, self.path)

        return self

    def exists(self) -> bool:
        if self.for_name:
            return self.spark_catalog.table_exists_from_path(self.path)
        else:
            return self.spark_fs_path.file_exists_from_path(self.path, False)

    def create_empty_if_not_exists(self, schema: StructType = None,
                                   partition_columns: list[str] = None) -> 'EasySparkDeltaTable':

        if not self.exists():
            self.create_empty(schema, partition_columns)

        return self

    def create_empty(self, schema: StructType = None, partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        easy_df = EasySparkDF.create(EasySparkDeltaTable._spark).empty(schema)

        self.save(easy_df, partition_columns=partition_columns)
        return self

    def get_dict(self, keys: dict[str, any]) -> dict[str, any] | None:
        df = self.to_easy_df().filter(keys).df
        if df is None or df.count() == 0:
            return None

        rows = df.collect()
        return rows[0].asDict()

    def get_dict_using_filter(self, keys: dict[str, any]) -> dict[str, any] | None:
        df = self.to_easy_df().filter_using_filter(keys).df
        if df is None or df.count() == 0:
            return None

        rows = df.collect()
        return rows[0].asDict()

    def get_dict_by_filter(self, condition: str) -> dict[str, any] | None:
        df = self.to_easy_df().filter_by_filter(condition).df
        if df is None or df.count() == 0:
            return None

        rows = df.collect()
        return rows[0].asDict()

    def get_dict_by_where(self, condition: str) -> dict[str, any] | None:
        df = self.to_easy_df().where(condition).df
        if df is None or df.count() == 0:
            return None

        rows = df.collect()
        return rows[0].asDict()

    def get_rows_using_filter(self, keys: dict[str, any] = None) -> list[Row]:
        return self.to_easy_df().filter_using_filter(keys).df.collect()

    def get_rows_by_filter(self, condition: str = None) -> list[Row]:
        return self.to_easy_df().filter_by_filter(condition).df.collect()

    def get_rows_by_where(self, condition: str = None) -> list[Row]:
        return self.to_easy_df().where(condition).df.collect()

    def get_rows_using_where(self, keys: dict[str, any] = None) -> list[Row]:
        return self.to_easy_df().where_using_where(keys).df.collect()

    def get_rows(self, keys: dict[str, any] = None) -> list[Row]:
        return self.to_easy_df().filter(keys).df.collect()

    def get_list(self, keys: dict[str, any] = None) -> list[dict[str, any]]:
        rows = self.get_rows(keys)
        return [row.asDict() for row in rows]

    def get_list_using_filter(self, keys: dict[str, any] = None) -> list[dict[str, any]]:
        rows = self.get_rows_using_filter(keys)
        return [row.asDict() for row in rows]

    def get_list_by_filter(self, condition: str = None) -> list[dict[str, any]]:
        rows = self.get_rows_by_filter(condition)
        return [row.asDict() for row in rows]

    def get_list_by_where(self, condition: str = None) -> list[dict[str, any]]:
        rows = self.get_rows_by_where(condition)
        return [row.asDict() for row in rows]

    def add_from_dict(self, record: dict, schema: StructType = None) -> 'EasySparkDeltaTable':
        return self.add_from_list([record], schema)

    def add_from_list(self, records: list[dict], schema: StructType = None,
                      ignore_order=False) -> 'EasySparkDeltaTable':
        new_easy_df = EasySparkDF.create(self.spark_instance.spark).clone_from_list(records, schema, ignore_order)
        self.save(new_easy_df, save_mode_type=EasySaveModeType.APPEND)
        return self

    def add_from_df(self, df: DataFrame, merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                    partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        new_easy_df = EasySparkDF.create(self.spark_instance.spark, df)
        self.save(new_easy_df, save_mode_type=EasySaveModeType.APPEND, merge_option_type=merge_option_type,
                  partition_columns=partition_columns)
        return self

    def combine_from_df(self, df: DataFrame, type: str = 'unionByName',
                        allow_missing_columns: bool = True,
                        merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                        partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        easy_df = self.to_easy_df()
        easy_df.combine_from_df(df, type, allow_missing_columns)
        self.save(easy_df, merge_option_type=merge_option_type, partition_columns=partition_columns)
        return self

    def add_from_pd_df(self, pd_df: DataFramePandas, schema: StructType = None,
                       merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                       partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        new_easy_df = EasySparkDF.create(self.spark_instance.spark).clone_from_pd_df(pd_df, schema)
        self.save(new_easy_df, save_mode_type=EasySaveModeType.APPEND, merge_option_type=merge_option_type,
                  partition_columns=partition_columns)
        return self

    def combine_from_pd_df(self, pd_df: DataFramePandas, schema: StructType = None,
                           type: str = 'unionByName',
                           allow_missing_columns: bool = True,
                           merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA) -> 'EasySparkDeltaTable':
        easy_df = self.to_easy_df()
        easy_df.combine_from_pd_df(pd_df, schema, type, allow_missing_columns)
        self.save(easy_df, merge_option_type=merge_option_type)
        return self

    def update(self, keys: dict[str, any], values: dict[str, any]) -> 'EasySparkDeltaTable':
        conditions = EasySparkDeltaTable._build_condition(keys)
        sets = {k: lit(v) for k, v in values.items()}

        self._delta_path.update(
            condition=conditions,
            set=sets
        )
        return self

    def update_by_condition(self, condition: str, values: dict[str, any]) -> 'EasySparkDeltaTable':
        sets = {k: lit(v) for k, v in values.items()}
        self._delta_path.update(
            condition=condition,
            set=sets
        )
        return self

    def delete(self, keys: dict[str, any] = None,
               multiple_keys: list[tuple[str, list]] = None) -> 'EasySparkDeltaTable':
        conditions = ""

        if keys:
            conditions = EasySparkDeltaTable._build_condition(keys)

        if multiple_keys and len(multiple_keys) > 0:
            conditions = EasySparkDeltaHelpers.build_condition_by_multiple_keys(multiple_keys, conditions)

        self._delta_path.delete(condition=conditions)
        return self

    def delete_by_multiple_keys(self, key: str, key_values: list) -> 'EasySparkDeltaTable':
        self._delta_path.delete(f"{key} in {tuple(key_values)}")
        return self

    def delete_by_condition(self, condition: str) -> 'EasySparkDeltaTable':
        self._delta_path.delete(condition)
        return self

    def delete_all(self) -> 'EasySparkDeltaTable':
        easy_df = self.to_easy_df()
        easy_df.empty()
        self.save(easy_df, save_mode_type=EasySaveModeType.OVER_WRITE)

        return self

    def save_from_df(self, df: DataFrame = None, save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                     merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                     partition_columns: list[str] = None):
        easy_df = EasySparkDF.create(EasySparkDeltaTable._spark, df)
        return self.save(easy_df, save_mode_type, merge_option_type, partition_columns)

    def save(self, easy_df: EasySparkDF = None, save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
             merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
             partition_columns: list[str] = None) -> 'EasySparkDeltaTable':
        if not easy_df:
            easy_df = self.to_easy_df()

        if self.for_name:
            easy_df.save_as_table(self.path, save_mode_type=save_mode_type, merge_option_type=merge_option_type,
                                  partition_columns=partition_columns)
        else:
            easy_df.save_to_table_from_path(self.path, save_mode_type=save_mode_type,
                                            merge_option_type=merge_option_type,
                                            partition_columns=partition_columns)

        return self

    def merge_from_list(self, merge_type: EasyMergeType, keys: list[str], records: list[dict],
                        schema: StructType = None,
                        add_missing_columns=True, add_missing_columns_to_current=False, ignore_order=False,
                        to_keys: dict[str, any] = None) -> 'EasySparkDeltaTable':

        if schema is None:
            easy_df = self.to_easy_df()
            schema = easy_df.current_schema

        if ignore_order:
            rows = [Row(**record) for record in records]
            df = EasySparkDeltaTable._spark.createDataFrame(rows, schema)
        else:
            pdf_df = pd.DataFrame(records)
            df = EasySparkDeltaTable._spark.createDataFrame(pdf_df, schema)

        return self.merge_from_df(merge_type, keys, df, add_missing_columns, add_missing_columns_to_current, to_keys)

    def merge_from_tuple(self, merge_type: EasyMergeType, keys: list[str], records: list[tuple],
                         schema: StructType = None,
                         add_missing_coloumns=True, add_missing_columns_to_current=False,
                         to_keys: dict[str, any] = None) -> 'EasySparkDeltaTable':
        if schema is None:
            easy_df = self.to_easy_df()
            schema = easy_df.current_schema

        df = EasySparkDeltaTable._spark.createDataFrame(records, schema)
        return self.merge_from_df(merge_type, keys, df, add_missing_coloumns, add_missing_columns_to_current, to_keys)

    def merge_from_df(self, merge_type: EasyMergeType, keys: list[str], df: DataFrame, add_missing_coloumns=True,
                      add_missing_columns_to_current=False, to_keys: dict[str, any] = None) -> 'EasySparkDeltaTable':
        current_easy_df = self.to_easy_df()
        current_df = current_easy_df.df
        df_columns = df.columns
        current_columns = current_df.columns

        if add_missing_coloumns:
            for current_column in current_columns:
                if current_column not in df_columns:
                    df = df.withColumn(current_column, lit(None).cast(current_df.schema[current_column].dataType))

        if add_missing_columns_to_current:
            current_df_has_new_columns = False
            for df_column in df_columns:
                if df_column not in current_columns:
                    current_df = current_df.withColumn(df_column, lit(None).cast(df.schema[df_column].dataType))
                    current_df_has_new_columns = True

            if current_df_has_new_columns:
                self.save_from_df(current_df, save_mode_type=EasySaveModeType.OVER_WRITE,
                                  merge_option_type=EasyMergeOptionType.OVERWRITE_SCHEMA)
                self.init_delta()

        merge_relationships = [f"A.`{key}` = B.`{key}` and " for key in keys]
        merge_relationships = "".join(merge_relationships)[:-4]

        if to_keys:
            to_conditions = EasySparkDeltaHelpers.build_condition(to_keys, "", "A")
            merge_relationships += f"And {to_conditions}"

        merge_entry = self._delta_path.alias('A').merge(
            df.alias('B'),
            merge_relationships
        )

        if merge_type == EasyMergeType.UPDATE_AND_INSERT:
            merge_entry = merge_entry.whenMatchedUpdateAll().whenNotMatchedInsertAll()
        elif merge_type == EasyMergeType.UPDATE_ONLY:
            merge_entry = merge_entry.whenMatchedUpdateAll()
        elif merge_type == EasyMergeType.INSERT_ONLY:
            merge_entry = merge_entry.whenNotMatchedInsertAll()

        merge_entry.execute()

        return self

    def optimize(self):
        if self.for_name:
            self.df_optimize.optimize_from_lh_table(self.path)
        else:
            self.df_optimize.optimize_from_path(self.path)

    def vacuum(self):
        if self.for_name:
            self.df_optimize.vacuum_from_lh_table(self.path)
        else:
            self.df_optimize.vacuum_from_path(self.path)

    # Private Method
    @staticmethod
    def _build_condition(keys: dict[str, any]):
        return EasySparkDeltaHelpers.build_condition(keys)



from pyspark.sql import Row
from pyspark.sql.functions import *
from pyspark.sql.types import StructField
from easy_panda.easy_panda import EasyPanda
from easy_spark.easy_format_type import EasyFormatType
from easy_spark.easy_save_type import EasySaveType
from easy_spark.easy_merge_option_type import EasyMergeOptionType
from easy_spark.easy_save_mode_type import EasySaveModeType
from easy_spark.easy_spark_delta_helpers import EasySparkDeltaHelpers
from easy_spark.easy_spark_df_optimize import EasySparkDFOptimize
from easy_spark.easy_spark_instance import EasySparkInstance
from easy_spark.easy_sql_builder import EasySQLSelectBuilder
from easy_spark_path.easy_file_type import EasyFileType
from easy_spark_path.easy_spark_directory_path import EasySparkDirectoryPath
from easy_spark_path.easy_spark_file_path import EasySparkFilePath
from easy_spark_path.easy_spark_table_path import EasySparkTablePath
from easy_spark_path.easy_spark_simple_table_path import EasySparkSimpleTablePath
from pyspark.sql import SparkSession
from datetime import datetime
import numpy as np
from pandas import DataFrame as DataFramePandas
import pandas as pd
from typing import Callable
from pyspark.sql.window import Window


class EasySparkDF:
    _spark: SparkSession = None

    # Constructor
    def __init__(self, df: DataFrame = None, spark: SparkSession = None):
        self._df: DataFrame = df
        EasySparkDF._spark = spark
        self.spark_instance = EasySparkInstance(spark)

    @staticmethod
    def create(spark: SparkSession, df: DataFrame = None) -> 'EasySparkDF':
        return EasySparkDF(df, spark)

    @staticmethod
    def create_empty(spark: SparkSession, schema: StructType = None) -> 'EasySparkDF':
        return EasySparkDF(None, spark).empty(schema)

    @property
    def df(self) -> DataFrame:
        return self._df

    @df.setter
    def df(self, value: DataFrame):
        self._df = value

    @property
    def current_schema(self) -> StructType:
        return self._df.schema if self._df and self._df.schema else None

    def from_simple_table_path(self, simple_table_path: EasySparkSimpleTablePath, df_format="delta",
                               options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        return self.from_path(simple_table_path.path, df_format, options)

    def from_table_path(self, table_path: EasySparkTablePath, df_format="delta",
                        options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        if table_path.is_table_name:
            return self.from_lh_table_path(table_path.path)
        else:
            return self.from_path(table_path.path, df_format, options)

    def from_path(self, path: str, df_format="delta", options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        entry = EasySparkDF._spark.read.format(df_format)
        if options:
            for option in options:
                entry = entry.option(option[0], option[1])
        self._df = entry.load(path)
        return self

    def from_lh_table_name(self, name: str, table_name: str, limit: int = None) -> 'EasySparkDF':
        sql_builder = EasySQLSelectBuilder.new().select(["*"]).from_lh_table_name(name, table_name).limit(limit)
        return self.from_sql(sql_builder.sql)

    def from_lh_table_path(self, path: str, limit: int = None) -> 'EasySparkDF':
        sql_builder = EasySQLSelectBuilder.new().select(["*"]).from_path(path).limit(limit)
        return self.from_sql(sql_builder.sql)

    def from_lh_table_name_using_sql(self, name: str, table_name: str, limit: int = None) -> 'EasySparkDF':
        if limit:
            self._df = EasySparkDF._spark.sql(f"SELECT * FROM {name}.{table_name} LIMIT {limit}")
        else:
            self._df = EasySparkDF._spark.sql(f"SELECT * FROM {name}.{table_name}")

        return self

    def from_sql_builder(self, sql_builder: EasySQLSelectBuilder) -> 'EasySparkDF':
        return self.from_sql(sql_builder.sql)

    def from_dict(self, records: dict, schema: StructType = None) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema
        self._df = EasySparkDF._spark.createDataFrame([Row(**records)], new_schema)
        return self

    def from_list(self, records: list[dict], schema: StructType = None, ignore_order=False) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema

        if ignore_order:
            rows = [Row(**record) for record in records]
            self._df = EasySparkDF._spark.createDataFrame(rows, new_schema)
        else:
            pdf_df = pd.DataFrame(records)
            self.from_pandas(pdf_df, new_schema)

        return self

    def from_tuple(self, records: list[tuple], schema: StructType = None) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema
        self._df = EasySparkDF._spark.createDataFrame(records, new_schema)
        return self

    def from_sql(self, sql: str) -> 'EasySparkDF':
        self._df = EasySparkDF._spark.sql(sql)
        return self

    def from_json(self, json: str) -> 'EasySparkDF':
        self._df = EasySparkDF._spark.read.json(json)
        return self

    def from_csv(self, csv_path: str, header=True, seperator=",", infer_schema=True) -> 'EasySparkDF':
        self._df = EasySparkDF._spark.read.option("header", header).csv(csv_path, sep=seperator,
                                                                        inferSchema=infer_schema)

        return self

    def from_file_path_csv(self, path: str, header=True, options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        entry = EasySparkDF._spark.read.format("csv")
        if header:
            entry = entry.option("header", "true")
        if options:
            for option in options:
                entry = entry.option(option[0], option[1])
        self._df = entry.load(path)

        return self

    def from_file_path_parquet(self, path: str, options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        entry = EasySparkDF._spark.read.format("parquet")
        if options:
            for option in options:
                entry = entry.option(option[0], option[1])

        self._df = entry.parquet(path)
        return self

    def from_file_path_json(self, path: str, multiline=True, options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        entry = EasySparkDF._spark.read.format("json")
        if multiline:
            entry = entry.option("multiline", "true")
        if options:
            for option in options:
                entry = entry.option(option[0], option[1])

        self._df = entry.json(path)
        return self

    def from_file_path(self, file_path: EasySparkFilePath, header=True, multiline=True,
                       options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        if file_path.file_type == EasyFileType.CSV:
            return self.from_file_path_csv(file_path.path, header, options)
        elif file_path.file_type == EasyFileType.PARQUET:
            return self.from_file_path_parquet(file_path.path, options)
        elif file_path.file_type == EasyFileType.JSON:
            return self.from_file_path_json(file_path.path, multiline, options)
        else:
            raise ValueError('File type not supported')

    def from_directory_path_using_wildcard(self, directory_path: EasySparkDirectoryPath, file_type: EasyFileType,
                                           header=True,
                                           multiline=True,
                                           options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        path = directory_path.path
        # Check if path ends with / if not add it
        if path[-1] != '/':
            path = path + '/'
        path = path + "*." + file_type.value

        if file_type == EasyFileType.CSV:
            return self.from_file_path_csv(path, header, options)
        elif file_type == EasyFileType.PARQUET:
            return self.from_file_path_parquet(path, options)
        elif file_type == EasyFileType.JSON:
            return self.from_file_path_json(path, multiline, options)
        else:
            raise ValueError('File type not supported')

    def from_easy_panda(self, easy_panda: EasyPanda, schema: StructType = None) -> 'EasySparkDF':
        return self.from_pandas(easy_panda.pd_df, schema)

    def from_pandas(self, pd_df: DataFramePandas, schema: StructType = None) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema
        self._df = EasySparkDF._spark.createDataFrame(pd_df, new_schema)

        return self

    def clear(self) -> 'EasySparkDF':
        self._df = EasySparkDF._spark.createDataFrame([], StructType([]))

        return self

    def empty(self, schema: StructType = None) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema
        self._df = EasySparkDF._spark.createDataFrame([], new_schema)

        return self

    def filter_by_filter(self, condition: str) -> 'EasySparkDF':
        self._df = self._df.filter(condition)
        return self

    def filter_by_any(self, condition: any) -> 'EasySparkDF':
        self._df = self._df.filter(condition)
        return self

    def filter(self, keys: dict[str, any] = None):
        for key in keys:
            self._df = self._df[self._df[key] == keys[key]]

        return self

    def filter_using_filter(self, keys: dict[str, any] = None):
        if keys:
            conditions = EasySparkDeltaHelpers.build_condition(keys)
            self._df = self._df.filter(conditions)

        return self

    def where_using_where(self, keys: dict[str, any] = None) -> 'EasySparkDF':
        if keys:
            conditions = EasySparkDeltaHelpers.build_condition(keys)
            self._df = self._df.where(conditions)
        return self

    def where(self, condition: str) -> 'EasySparkDF':
        if condition:
            self._df = self._df.where(condition)
        return self

    def where_by_any(self, condition: any) -> 'EasySparkDF':
        if condition:
            self._df = self._df.where(condition)
        return self

    def combine_from_df(self, df: DataFrame, type: str = 'unionByName',
                        allow_missing_columns: bool = True) -> 'EasySparkDF':
        self._df = EasySparkDeltaHelpers.combine_from_dfs([self._df, df], type, allow_missing_columns)
        return self

    def combine_from_pd_df(self, pd_df: DataFramePandas, schema: StructType = None, type: str = 'unionByName',
                           allow_missing_columns: bool = True) -> 'EasySparkDF':
        new_schema = schema if schema is not None else self.current_schema
        df = EasySparkDF._spark.createDataFrame(pd_df, new_schema)
        return self.combine_from_df(df, type, allow_missing_columns)

    def rename_column_empty_spaces(self, value="") -> 'EasySparkDF':
        for col_name in self._df.columns:
            self._df = self._df.withColumnRenamed(col_name, col_name.replace(" ", value))
        return self

    def rename_column_invalid_chars(self, invalid_chars=",;{}()", value="") -> 'EasySparkDF':
        for col_name in self._df.columns:
            current_col_name = col_name
            for c in invalid_chars:
                current_col_name = current_col_name.replace(c, value)
            self._df = self._df.withColumnRenamed(col_name, current_col_name)

        return self

    def rename_column_wildcard(self, from_value="", value="") -> 'EasySparkDF':
        for col_name in self._df.columns:
            current_col_name = col_name
            current_col_name = current_col_name.replace(from_value, value)
            self._df = self._df.withColumnRenamed(col_name, current_col_name)

        return self

    def rename_columns(self, values: dict[str, str]):
        for col_name in self._df.columns:
            if col_name in values:
                self._df = self._df.withColumnRenamed(col_name, values[col_name])

        return self

    def rename_column(self, from_value="", value=""):
        ##First check if the column exists
        if from_value in self._df.columns:
            self._df = self._df.withColumnRenamed(from_value, value)

        return self

    def drop_column_if_exists(self, column_name: str) -> 'EasySparkDF':
        if column_name in self._df.columns:
            self._df = self._df.drop(column_name)
        return self

    def add_column(self, column_name: str, value: any) -> 'EasySparkDF':
        self._df = self._df.withColumn(column_name, lit(value))
        return self

    def capitalize_column_between_spaces(self) -> 'EasySparkDF':
        for col_name in self._df.columns:
            current_col_name = col_name
            # Capitalize the first letter of each word
            for word in current_col_name.split(" "):
                current_col_name = current_col_name.replace(word, word.capitalize())
            self._df = self._df.withColumnRenamed(col_name, current_col_name)

        return self

    def title_columns(self) -> 'EasySparkDF':
        for col_name in self._df.columns:
            self._df = self._df.withColumnRenamed(col_name, col_name.title())

        return self

    def replace_nan(self, value=None) -> 'EasySparkDF':
        self._df = self._df.replace({np.nan: value}).replace({"nan": value})

        return self

    def overwrite_types(self, struct_types: dict | list[StructField] = None) -> 'EasySparkDF':
        struct_types = EasySparkDeltaHelpers.build_struct_fields(struct_types)

        for struct_t in struct_types:
            filed_r = struct_t.name.replace("`", "")
            # Check if field is in df
            if filed_r not in self._df.columns:
                continue
            self._df = self._df.withColumn(filed_r, col(struct_t.name).cast(struct_t.dataType))

        return self

    def overwrite_to_dates(self, overwrite_columns: list[tuple[str, str]]) -> 'EasySparkDF':
        for (column_name, column_date_format) in overwrite_columns:
            if column_date_format:
                self._df = self._df.withColumn(column_name,
                                               to_timestamp(self._df[column_name], column_date_format))
            else:
                self._df = self._df.withColumn(column_name, to_timestamp(self._df[column_name]))

        return self

    def add_audit_fields(self) -> 'EasySparkDF':
        now_date = datetime.now()
        date_only = now_date.date()
        self._df = self._df.withColumn("CreatedDate", lit(now_date))
        self._df = self._df.withColumn("CreatedYear", lit(now_date.year))
        self._df = self._df.withColumn("CreatedMonth", lit(now_date.month))
        self._df = self._df.withColumn("CreatedDay", lit(now_date.day))
        self._df = self._df.withColumn("CreatedDateOnly", lit(date_only))

        return self

    def remove_audit_fields(self) -> 'EasySparkDF':
        self.drop_column_if_exists("CreatedDate").drop_column_if_exists("CreatedYear").drop_column_if_exists(
            "CreatedMonth").drop_column_if_exists("CreatedDay").drop_column_if_exists("CreatedDateOnly")

        return self

    def remove_audit_field_columns(self):
        return self.remove_audit_fields()

    def add_hash_column_from_columns(self, hash_column_name: str, columns: list[str] = None) -> 'EasySparkDF':
        if not columns:
            columns = self._df.columns

        self._df = self._df.withColumn(hash_column_name, sha2(concat_ws("", *columns), 256))
        return self

    def append_from_dict(self, record: dict) -> 'EasySparkDF':
        row = Row(**record)
        return self.append_from_row(row)

    def append_from_row(self, row: Row) -> 'EasySparkDF':
        df = EasySparkDF._spark.createDataFrame([row], self._df.schema)
        self._df = self.combine_from_df(df, type='union', allow_missing_columns=True)._df
        return self

    def select(self, columns: list[str]) -> 'EasySparkDF':
        self._df = self._df.select(columns)
        return self

    def select_except(self, columns: list[str]) -> 'EasySparkDF':
        self._df = self._df.select([col for col in self._df.columns if col not in columns])
        return self

    def clone(self, new_schema: StructType = None) -> 'EasySparkDF':
        if new_schema:
            new_df = EasySparkDF._spark.createDataFrame(self._df.rdd, new_schema)
            return EasySparkDF.create(EasySparkDF._spark, new_df)
        else:
            new_df = EasySparkDF._spark.createDataFrame(self._df.rdd, self.current_schema)
            return EasySparkDF.create(EasySparkDF._spark, new_df)

    def clone_from_pd_df(self, pd_df: DataFramePandas, schema: StructType = None) -> 'EasySparkDF':
        return EasySparkDF.create(EasySparkDF._spark).from_pandas(pd_df, schema)

    def clone_from_list(self, records: list[dict], schema: StructType = None, ignore_order=False) -> 'EasySparkDF':
        return EasySparkDF.create(EasySparkDF._spark).from_list(records, schema, ignore_order)

    def copy(self, new_schema: StructType = None) -> 'EasySparkDF':
        if new_schema:
            new_df = self._df
            new_df = EasySparkDeltaHelpers.remove_columns_based_on_schema(new_df, new_schema)
            new_df = EasySparkDeltaHelpers.add_missing_columns_based_on_schema(new_df, new_schema)
            new_df = EasySparkDF._spark.createDataFrame(new_df.rdd, new_schema)

            return EasySparkDF.create(EasySparkDF._spark, new_df)
        else:
            return EasySparkDF.create(EasySparkDF._spark, self._df)

    def distinct(self, columns: list[str] = None) -> 'EasySparkDF':
        if columns:
            self._df = self._df.distinct(columns)
        else:
            self._df = self._df.distinct()
        return self

    def order_columns(self, columns: list[str]) -> 'EasySparkDF':
        # Order the columns in the order of the list
        self._df = self._df.select(columns)
        return self

    def add_partitions(self, columns: list[str]) -> 'EasySparkDF':
        self._df = self._df.repartition(*columns)
        return self

    def is_empty_or_null(self) -> bool:
        return self._df is None or self._df.isEmpty()

    def delete_empty_rows(self, columns: list[str]) -> 'EasySparkDF':
        self._df = self._df.dropna(subset=columns)
        return self

    def collect_distinct_values(self, column_name: str) -> list[any]:
        return self._df.select(column_name).distinct().rdd.map(lambda x: x[0]).collect()

    def collect_distinct_rows(self, columns: list[str]) -> list[Row]:
        return self._df.select(columns).distinct().collect()

    def collect(self) -> list[Row]:
        return self._df.collect()

    def iterator(self):
        return self._df.toLocalIterator()

    def transform(self, func: Callable[[DataFrame], DataFrame]) -> 'EasySparkDF':
        self._df = func(self._df)
        return self

    def delete_duplicates_by_cols(self, columns: list[str],
                                  *cols: Union["ColumnOrName", List["ColumnOrName_"]]) -> 'EasySparkDF':
        window_spec = Window.partitionBy(columns).orderBy(*cols)
        self._df = self._df.withColumn("row_num", row_number().over(window_spec))
        self._df = self._df.filter(self._df.row_num == 1)
        self._df = self._df.drop("row_num")
        return self

    def delete_duplicates_by_desc(self, columns: list[str], column_name: str) -> 'EasySparkDF':
        return self.delete_duplicates_by_cols(columns, col(column_name).desc())

    def save_to_table_from_table_path(self, table_path: EasySparkTablePath,
                                      save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                                      merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                                      partition_columns: list[str] = None,
                                      options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        if table_path.is_table_name:
            return self.save(table_path.path, EasySaveType.TABLE, EasyFormatType.DELTA, save_mode_type,
                             merge_option_type,
                             partition_columns, options)
        else:
            return self.save(table_path.path, EasySaveType.SAVE, EasyFormatType.DELTA, save_mode_type,
                             merge_option_type,
                             partition_columns, options)

    def save_to_table_from_simple_table_path(self, simple_table_path: EasySparkSimpleTablePath,
                                             save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                                             merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                                             partition_columns: list[str] = None,
                                             options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        return self.save(simple_table_path.path, EasySaveType.SAVE, EasyFormatType.DELTA, save_mode_type,
                         merge_option_type,
                         partition_columns,
                         options)

    def save_as_table(self, path: str, save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                      merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                      partition_columns: list[str] = None,
                      options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        return self.save(path, EasySaveType.TABLE, EasyFormatType.DELTA, save_mode_type, merge_option_type,
                         partition_columns, options)

    def save_to_table_from_path(self, path: str, save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                                merge_option_type: EasyMergeOptionType = EasyMergeOptionType.OVERWRITE_SCHEMA,
                                partition_columns: list[str] = None,
                                options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        return self.save(path, EasySaveType.SAVE, EasyFormatType.DELTA, save_mode_type, merge_option_type,
                         partition_columns, options)

    def save_from_file_path(self, file_path: EasySparkFilePath,
                            save_mode_type: EasySaveModeType = EasySaveModeType.OVER_WRITE,
                            merge_option_type: EasyMergeOptionType = None, partition_columns: list[str] = None,
                            options: list[tuple[str, any]] = None):
        if file_path.file_type == EasyFileType.PARQUET:
            return self.save(file_path.path, EasySaveType.PARQUET, None, save_mode_type,
                             merge_option_type, partition_columns,
                             options)
        elif file_path.file_type == EasyFileType.CSV:
            return self.save(file_path.path, EasySaveType.CSV, None, save_mode_type, merge_option_type,
                             partition_columns,
                             options)
        elif file_path.file_type == EasyFileType.JSON:
            return self.save(file_path.path, EasySaveType.JSON, None, save_mode_type, merge_option_type,
                             partition_columns,
                             options)
        else:
            raise ValueError('File type not supported')

    def save(self, path: str, save_type: EasySaveType, format_type: EasyFormatType = None,
             save_mode_type: EasySaveModeType | None = EasySaveModeType.OVER_WRITE,
             merge_option_type: EasyMergeOptionType = None, partition_columns: list[str] = None,
             options: list[tuple[str, any]] = None) -> 'EasySparkDF':
        entry = self._df.write
        if format_type:
            entry = entry.format(format_type.value)
        if save_mode_type:
            entry = entry.mode(save_mode_type.value)
        if merge_option_type:
            entry = entry.option(merge_option_type.value, "true")
        if options:
            for option in options:
                entry = entry.option(option[0], option[1])
        if partition_columns:
            entry = entry.partitionBy(partition_columns)
        if save_type == EasySaveType.TABLE:
            entry.saveAsTable(path)
        elif save_type == EasySaveType.SAVE:
            entry.save(path)
        elif save_type == EasySaveType.PARQUET:
            entry.parquet(path)
        elif save_type == EasySaveType.CSV:
            entry.csv(path)
        elif save_type == EasySaveType.JSON:
            entry.json(path)

        return self



class EasySQLDataConnectionBuilder:
    def __init__(self, constr: str):
        self.constr = constr

    @staticmethod
    def init_using_credentials(server: str, database: str, username: str, password: str):
        constr = (
            f"Driver={{ODBC Driver 18 for SQL Server}};"
            f"Server={server};"
            f"Database={database};"
            f"UID={username};"
            f"PWD={password};"
            f"Authentication=ActiveDirectoryPassword;"
            f"Encrypt=yes;"
            f"TrustServerCertificate=no;"
            f"Connection Timeout=30;"
        )
        return EasySQLDataConnectionBuilder(constr)



from pyspark import SparkContext
from pyspark.sql import SparkSession
from py4j.java_gateway import java_import
from easy_spark.easy_spark_instance import EasySparkInstance
from easy_utils.easy_singleton import easy_singleton


@easy_singleton
class EasySparkHadoop:
    def __init__(self, spark: SparkSession = None, must_java_import=True):
        self.spark_instance = EasySparkInstance(spark)
        self.hadoop_conf = None
        self.fs = None
        self.sc: SparkContext = None
        self.make_path = None

        if must_java_import:
            self.import_java()

        self.configure_fs()
        pass

    def import_java(self):
        spark = self.spark_instance.spark
        java_import(spark._jvm, 'org.apache.hadoop.fs.FileSystem')
        java_import(spark._jvm, 'org.apache.hadoop.conf.Configuration')
        java_import(spark._jvm, 'org.apache.hadoop.fs.Path')

    def configure_fs(self):
        spark = self.spark_instance.spark
        self.sc = spark.sparkContext
        self.hadoop_conf = self.sc._jsc.hadoopConfiguration()
        self.fs = self.sc._jvm.org.apache.hadoop.fs.FileSystem.get(self.hadoop_conf)
        self.make_path = self.sc._jvm.org.apache.hadoop.fs.Path



from easy_spark.easy_spark_delta_helpers import EasySparkDeltaHelpers


class EasySQLSelectBuilder:
    def __init__(self):
        self.sql: str = "SELECT"
        pass;

    def select(self, columns: list[str]) -> 'EasySQLSelectBuilder':
        self.sql += " " + ", ".join(columns)
        return self

    def from_table_name(self, table: str) -> 'EasySQLSelectBuilder':
        self.sql += f" FROM {table}"
        return self

    def from_lh_table_name(self, name: str, table_name: str) -> 'EasySQLSelectBuilder':
        self.sql += f" FROM {name}.{table_name}"
        return self

    def from_path(self, path: str) -> 'EasySQLSelectBuilder':
        self.sql += f" FROM delta.`{path}`"
        return self

    def where(self, keys: dict[str, any]) -> 'EasySQLSelectBuilder':
        conditions = EasySparkDeltaHelpers.build_condition(keys)

        self.sql += f" WHERE {conditions}"
        return self

    def where_from_condition(self, condition: str) -> 'EasySQLSelectBuilder':
        self.sql += f" WHERE {condition}"
        return self

    def limit(self, limit: int) -> 'EasySQLSelectBuilder':
        if limit:
            self.sql += f" LIMIT {limit}"
        return self

    @staticmethod
    def new() -> 'EasySQLSelectBuilder':
        return EasySQLSelectBuilder()



from easy_spark.easy_spark_instance import EasySparkInstance
from pyspark.sql import SparkSession


class EasySparkConf:
    def __init__(self, spark: SparkSession = None):
        self.spark_instance = EasySparkInstance(spark)
        pass

    @staticmethod
    def create(spark: SparkSession = None):
        return EasySparkConf(spark)

    def optimize(self, optimize_write_enabled=True, rebase_mode_in_write=True, rebase_mode_in_read=True,
                 vorder_enabled: bool = False, low_shuffle_enabled: bool = False, concurrent_writes: bool = False,
                 ignore_corrupt_files: bool = False, auto_merge: bool = False, optimize_write_binsize=None):
        if rebase_mode_in_write:
            self.spark_instance.spark.conf.set("spark.sql.parquet.int96RebaseModeInWrite", "CORRECTED")
            self.spark_instance.spark.conf.set("spark.sql.parquet.datetimeRebaseModeInWrite", "CORRECTED")
        if rebase_mode_in_read:
            self.spark_instance.spark.conf.set("spark.sql.parquet.int96RebaseModeInRead", "CORRECTED")
            self.spark_instance.spark.conf.set("spark.sql.parquet.datetimeRebaseModeInRead", "CORRECTED")
        if optimize_write_enabled:
            self.spark_instance.spark.conf.set("spark.microsoft.delta.optimizeWrite.enabled", "true")
        if vorder_enabled:
            self.spark_instance.spark.conf.set("spark.sql.parquet.vorder.enabled", "true")
        if low_shuffle_enabled:
            self.spark_instance.spark.conf.set("spark.microsoft.delta.merge.lowShuffle.enabled", "true")
        if ignore_corrupt_files:
            self.spark_instance.spark.conf.set("spark.sql.files.ignoreCorruptFiles", "true")
        if auto_merge:
            self.spark_instance.spark.conf.set("spark.databricks.delta.schema.autoMerge.enabled", "true")
        if concurrent_writes:
            self.spark_instance.spark.conf.set("spark.databricks.delta.concurrentWrites", "true")
        if optimize_write_binsize:
            self.spark_instance.spark.conf.set("spark.microsoft.delta.optimizeWrite.binSize", optimize_write_binsize)
            self.spark_instance.spark.conf.set("spark.microsoft.delta.optimizedWrite.binSize", optimize_write_binsize)



from easy_spark_path.easy_spark_lh_path import EasySparkLHPath
from easy_spark_path.easy_path import Path
from easy_spark_path.easy_spark_simple_table_path import EasySparkSimpleTablePath
from easy_spark_path.easy_warehouse_type import EasyWarehouseType
from easy_spark_path.easy_spark_wh_path import EasySparkWHPath


class EasySparkTablePath:
    def __init__(self, path: str, is_table_name: bool, warehouse_type: EasyWarehouseType):
        self.path = path
        self.is_table_name = is_table_name
        self.warehouse_type = warehouse_type

    @staticmethod
    def create_from_lh_table_name(lh_name: str, table_name, schema: str = None):
        path = f'{lh_name}.{schema}.{table_name}' if schema is not None else f'{lh_name}.{table_name}'
        return EasySparkTablePath(path, True, EasyWarehouseType.LH)

    @staticmethod
    def create_from_wh_table_name(wh_name: str, table_name, schema: str):
        path = f'{wh_name}.{schema}.{table_name}'
        return EasySparkTablePath(path, True, EasyWarehouseType.WH)

    @staticmethod
    def create_from_simple_table_path(simple_table_path: EasySparkSimpleTablePath) -> 'EasySparkTablePath':
        warehouse_type = EasyWarehouseType.LH if simple_table_path.is_lh else EasyWarehouseType.WH
        return EasySparkTablePath(simple_table_path.path, False, warehouse_type)

    @staticmethod
    def create_from_lh_path(lh_table_path: EasySparkLHPath) -> 'EasySparkTablePath':
        return EasySparkTablePath.create_from_simple_table_path(lh_table_path)

    @staticmethod
    def create_from_wh_path(wh_table_path: EasySparkWHPath) -> 'EasySparkTablePath':
        return EasySparkTablePath.create_from_simple_table_path(wh_table_path)

    @staticmethod
    def create_from_path(path: Path, schema: str, table_name: str, is_lh) -> 'EasySparkTablePath':
        return EasySparkTablePath.create_from_simple_table_path(
            EasySparkSimpleTablePath(path, schema, table_name, is_lh))

    @staticmethod
    def create_from_path_details(workspace_id: str, warehouse_id: str, schema: str, table_name: str,
                                 is_lh) -> 'EasySparkTablePath':
        return EasySparkTablePath.create_from_simple_table_path(
            EasySparkSimpleTablePath(Path(workspace_id, warehouse_id), schema, table_name, is_lh))



from pyspark.sql import SparkSession
from datetime import datetime
from easy_spark_fs.easy_spark_fs_path import EasySparkFSPath
from easy_spark_fs.easy_spark_hoop import EasySparkHadoop
from easy_spark_path.easy_spark_file_path import EasySparkFilePath


class EasySparkFSFile:
    def __init__(self, spark: SparkSession = None, path: str = None, must_java_import=True, is_relative: bool = False):
        self.easy_spark_hadoop = EasySparkHadoop(spark, must_java_import)
        self.easy_spark_fs_path = EasySparkFSPath(spark)
        self.fs_path = None
        self.path: str = path
        self.file_status = None

        if self.path:
            self.from_path(self.path, is_relative)

    @staticmethod
    def create(spark: SparkSession = None, path: str = None, must_java_import=False,
               is_relative: bool = False) -> 'EasySparkFSFile':
        return EasySparkFSFile(spark, path, must_java_import, is_relative)

    @staticmethod
    def create_from_file_path(spark: SparkSession = None, file_path: EasySparkFilePath = None,
                              must_java_import=False) -> 'EasySparkFSFile':
        return EasySparkFSFile(spark, file_path.path, must_java_import, file_path.is_relative)

    def from_path(self, path: str, is_relative: bool = False) -> 'EasySparkFSFile':
        self.path = path
        self.fs_path = self.easy_spark_fs_path.get_fs_path(self.path, is_relative)
        self.init_file_status()
        return self

    def init_file_status(self) -> 'EasySparkFSFile':
        if self.fs_path and self.file_exists():
            self.file_status = self.easy_spark_fs_path.get_file_status_from_fs_path(self.fs_path)
        return self

    def file_exists(self) -> bool:
        return self.easy_spark_fs_path.file_exists_from_fs_path(self.fs_path)

    def get_file_status(self):
        return self.file_status

    def get_modified_time(self):
        return self.easy_spark_fs_path.get_file_modification_time_from_status(self.file_status)

    def get_readable_modified_time(self) -> datetime:
        return self.easy_spark_fs_path.get_readable_file_modification_time_from_status(self.file_status)

    def get_name(self) -> str:
        return self.easy_spark_fs_path.get_file_name_from_status(self.file_status)

    def get_full_path(self) -> str:
        return self.easy_spark_fs_path.get_file_full_path_from_status(self.file_status)

    def delete_file(self) -> bool:
        return self.easy_spark_fs_path.delete_file_from_fs_path(self.fs_path)

    def write_file_content(self, content: any, delete_if_exists: bool = False) -> bool:
        if delete_if_exists:
            if self.file_exists():
                self.delete_file()
        elif self.file_exists():
            return False

        output_stream = self.easy_spark_hadoop.fs.create(self.fs_path)
        try:
            output_stream.write(content)
        finally:
            output_stream.close()

        return True



from easy_sql_data_access.easy_sql_data_access import EasySQLDataAccess
from easy_sql_data_access.easy_sql_data_access_factory import EasySQLDataAccessFactory
import threading


class EasySQLDataEntityDict:
    def __init__(self, data_access: EasySQLDataAccessFactory | EasySQLDataAccess,
                 table_name: str, id_column_name: str = "Id", add_lock=False):
        self.data_access = data_access
        self.table_name = table_name
        self.id_column_name = id_column_name
        self.add_lock = add_lock
        self.lock = threading.Lock()
        pass

    def update(self, id: any, entity: dict[str, any]):
        def perform_update(data_access_inner: EasySQLDataAccess):
            if self.add_lock:
                with self.lock:
                    data_access_inner.update(id, entity)
            else:
                data_access_inner.update(id, entity)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                perform_update(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            perform_update(self.data_access)

        pass

    def patch(self, id: any, entity: dict[str, any]):
        def perform_patch(data_access_inner: EasySQLDataAccess):
            if self.add_lock:
                with self.lock:
                    data_access_inner.patch(id, entity)
            else:
                data_access_inner.patch(id, entity)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                perform_patch(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            perform_patch(self.data_access)

        pass

    def insert(self, entity: dict[str, any]):
        def perform_insert(data_access_inner: EasySQLDataAccess):
            if self.add_lock:
                with self.lock:
                    data_access_inner.insert(entity)
            else:
                data_access_inner.insert(entity)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                perform_insert(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            perform_insert(self.data_access)

        pass

    def delete(self, id: any):
        def perform_delete(data_access_inner: EasySQLDataAccess):
            if self.add_lock:
                with self.lock:
                    data_access_inner.delete(id)
            else:
                data_access_inner.delete(id)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                perform_delete(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            perform_delete(self.data_access)

        pass

    def get_list(self) -> list[dict[str, any]]:
        def perform_get_list(data_access_inner: EasySQLDataAccess) -> list[dict[str, any]]:
            if self.add_lock:
                with self.lock:
                    return data_access_inner.get_list()
            else:
                return data_access_inner.get_list()

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                return perform_get_list(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            return perform_get_list(self.data_access)

    def get_list_with_filters(self, filters: dict[str, any]) -> list[dict[str, any]]:
        def perform_get_list_with_filters(data_access_inner: EasySQLDataAccess) -> list[dict[str, any]]:
            if self.add_lock:
                with self.lock:
                    return data_access_inner.get_list_with_filters(filters)
            else:
                return data_access_inner.get_list_with_filters(filters)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                return perform_get_list_with_filters(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            return perform_get_list_with_filters(self.data_access)

    def get(self, id: any) -> dict[str, any]:
        def perform_get(data_access_inner: EasySQLDataAccess) -> dict[str, any]:
            if self.add_lock:
                with self.lock:
                    return data_access_inner.get(id)
            else:
                return data_access_inner.get(id)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                return perform_get(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            return perform_get(self.data_access)

        pass

    def get_with_filters(self, filters: dict[str, any]) -> dict[str, any]:
        def perform_get_with_filters(data_access_inner: EasySQLDataAccess) -> dict[str, any]:
            if self.add_lock:
                with self.lock:
                    return data_access_inner.get_with_filters(filters)
            else:
                return data_access_inner.get_with_filters(filters)

        if isinstance(self.data_access, EasySQLDataAccessFactory):
            with self.data_access.open() as data_access:
                return perform_get_with_filters(data_access)
        elif isinstance(self.data_access, EasySQLDataAccess):
            return perform_get_with_filters(self.data_access)

        pass



from pyspark.sql import SparkSession
from easy_utils.easy_singleton import easy_singleton


@easy_singleton
class EasySparkInstance:

    def __init__(self, spark: SparkSession = None):
        if spark:
            self._spark = spark
        pass

    @property
    def spark(self) -> SparkSession:
        return self._spark



import os


class EasyOS:
    @staticmethod
    def get_file_name_from_path(path: str) -> str:
        return os.path.basename(path)

    @staticmethod
    def delete_directory(path: str) -> None:
        os.rmdir(path)

    @staticmethod
    def delete_directory_if_exists(path: str) -> None:
        if os.path.exists(path):
            EasyOS.delete_directory(path)

    @staticmethod
    def delete_file(path: str) -> None:
        os.remove(path)

    @staticmethod
    def delete_file_if_exists(path: str) -> None:
        if os.path.exists(path):
            EasyOS.delete_file(path)

    @staticmethod
    def create_directory(path: str, exist_ok=True) -> None:
        os.makedirs(path, exist_ok=exist_ok)

    @staticmethod
    def create_directory_if_not_exists(path: str) -> None:
        if not os.path.exists(path):
            EasyOS.create_directory(path)

    @staticmethod
    def get_file_extension_from_path(path: str) -> str:
        return os.path.splitext(path)[1]

    @staticmethod
    def get_file_name_without_extension(path: str) -> str:
        return os.path.splitext(os.path.basename(path))[0]

    @staticmethod
    def get_file_name_and_extension(path: str) -> tuple[str, str]:
        return os.path.splitext(os.path.basename(path))

    @staticmethod
    def get_parent_directory(path: str) -> str:
        return os.path.dirname(path)

    @staticmethod
    def get_parent_directory_name(path: str) -> str:
        return os.path.basename(os.path.dirname(path))

    @staticmethod
    def get_parent_directory_path(path: str) -> str:
        return os.path.dirname(path)

    @staticmethod
    def get_parent_directory_name_and_path(path: str) -> tuple[str, str]:
        return os.path.basename(os.path.dirname(path)), os.path.dirname(path)



import pandas as pd
from pandas import DataFrame as DataFramePandas


class EasyPandaHelpers:
    @staticmethod
    def order_pd_types_to_pd_df(pd_df: DataFramePandas, pd_types: dict[str, str]) -> dict[str, str]:
        if pd_df.empty:
            return pd_types

        pd_types = {key: pd_types[key] for key in pd_df.columns}
        return pd_types

    @staticmethod
    def combine_pd_types(pd_types: dict[str, str], pd_types_to_add: dict[str, str]) -> dict[
        str, str]:
        for key, value in pd_types_to_add.items():
            if key not in pd_types:
                pd_types[key] = value
        return pd_types

    @staticmethod
    def overwrite_pd_types(pd_types: dict[str, str], pd_types_to_overwrite: dict[str, str]) -> dict[
        str, str]:
        for key, value in pd_types_to_overwrite.items():
            if key in pd_types:
                pd_types[key] = value
        return pd_types



class Path:
    def __init__(self, workspace_id: str, warehouse_id: str):
        self.workspace_id = workspace_id
        self.warehouse_id = warehouse_id
        self.path = f"abfss://{workspace_id}@onelake.dfs.fabric.microsoft.com/{warehouse_id}"



class EasySQLDataAccess:
    def __init__(self, con):
        self.con = con

    @staticmethod
    def create(con):
        return EasySQLDataAccess(con)

    def new_cursor(self):
        if self.con is None:
            raise ValueError("Connection is not open")
        return self.con.cursor()

    @staticmethod
    def execute_cursor(cursor, sql_command: str) -> any:
        result = cursor.execute(sql_command)
        return result

    @staticmethod
    def execute_cursor_with_parameters(cursor, sql_command: str, parameters: tuple) -> any:
        if parameters:
            return cursor.execute(sql_command, parameters)
        else:
            return cursor.execute(sql_command)

    def execute(self, sql_command: str):
        cursor = self.new_cursor()
        try:
            self.execute_cursor(cursor, sql_command)
        finally:
            cursor.close()

    def execute_with_parameters(self, sql_command: str, parameters: tuple):
        cursor = self.new_cursor()
        try:
            self.execute_cursor_with_parameters(cursor, sql_command, parameters)
        finally:
            cursor.close()

    def query_list_dict(self, sql_command: str, parameters: tuple = None) -> list[dict[str, any]]:
        print('Querying...', sql_command)
        cursor = self.new_cursor()
        try:
            result = self.execute_cursor_with_parameters(cursor, sql_command, parameters)
            list_result = [{column[0]: value for column, value in zip(result.description, row)} for row in
                           result.fetchall()]
            return list_result
        finally:
            cursor.close()

    def query_dict(self, sql_command: str, parameters: tuple = None) -> dict[str, any]:
        cursor = self.new_cursor()
        try:
            result = self.execute_cursor_with_parameters(cursor, sql_command, parameters)
            return {column[0]: value for column, value in zip(result.description, result.fetchone())}
        finally:
            cursor.close()

    def query_list_tuple(self, sql_command: str, parameters: tuple = None) -> list[tuple]:
        cursor = self.new_cursor()
        try:
            result = self.execute_cursor_with_parameters(cursor, sql_command, parameters)
            return result.fetchall()
        finally:
            cursor.close()

    def query_tuple(self, sql_command: str, parameters: tuple = None) -> tuple:
        cursor = self.new_cursor()
        try:
            result = self.execute_cursor_with_parameters(cursor, sql_command, parameters)
            return result.fetchone()
        finally:
            cursor.close()

    def update(self, id: any, entity: dict[str, any]):
        sql = f"UPDATE {self.table_name} SET " + ', '.join(
            [f"{key} = ?" for key in entity.keys()]) + f" WHERE {self.id_column_name} = ?"
        self.execute_with_parameters(sql, tuple(entity.values()) + (id,))

        pass

    def patch(self, id: any, entity: dict[str, any]):
        sql = f"UPDATE {self.table_name} SET " + ', '.join(
            [f"{key} = ?" for key in entity.keys()]) + f" WHERE {self.id_column_name} = ?"
        self.execute_with_parameters(sql, tuple(entity.values()) + (id,))

        pass

    def insert(self, entity: dict[str, any]):
        sql = f"INSERT INTO {self.table_name} ({', '.join(entity.keys())}) VALUES ({', '.join(['?' for _ in entity.keys()])})"
        self.execute_with_parameters(sql, tuple(entity.values()))

        pass

    def delete(self, id: any):
        sql = f"DELETE FROM {self.table_name} WHERE {self.id_column_name} = ?"
        self.execute_with_parameters(sql, (id,))

        pass

    def get_list(self) -> list[dict[str, any]]:
        sql = f"SELECT * FROM {self.table_name}"
        dict_list = self.query_list_dict(sql)
        return dict_list

    def get_list_with_filters(self, filters: dict[str, any]) -> list[dict[str, any]]:
        sql = f"SELECT * FROM {self.table_name} WHERE " + ' AND '.join([f"{key} = ?" for key in filters.keys()])

        dict_list = self.query_list_dict(sql, tuple(filters.values()))
        return dict_list

    def get(self, id: any) -> dict[str, any]:
        sql = f"SELECT * FROM {self.table_name} WHERE {self.id_column_name} = ?"
        dict_record = self.query_dict(sql, (id,))
        return dict_record

        pass

    def get_with_filters(self, filters: dict[str, any]) -> dict[str, any]:
        sql = f"SELECT * FROM {self.table_name} WHERE " + ' AND '.join([f"{key} = ?" for key in filters.keys()])

        dict_record = self.query_dict(sql, tuple(filters.values()))
        return dict_record

        pass



from easy_spark_path.easy_path import Path
from easy_spark_path.easy_spark_simple_table_path import EasySparkSimpleTablePath


class EasySparkLHPath(EasySparkSimpleTablePath):
    def __init__(self, path: Path, schema: str, table_name: str):
        super().__init__(path, schema, table_name, is_lh=True)



from easy_spark_fs.easy_file_detail import EasyFileDetails
from easy_spark_fs.easy_spark_hoop import EasySparkHadoop
from easy_spark_path.easy_spark_file_path import EasySparkFilePath
from easy_utils.easy_singleton import easy_singleton
from datetime import datetime
from typing import Iterator


@easy_singleton
class EasySparkFSPath:
    def __init__(self, spark, must_java_import=True):
        self.easy_hadoop = EasySparkHadoop(spark, must_java_import)

    def get_fs_path_from_status(self, status):
        return status.getPath()

    def get_file_name_from_status(self, status) -> str:
        return status.getPath().getName()

    def get_file_full_path_from_status(self, status) -> str:
        return status.getPath().toString()

    def get_file_modification_time_from_status(self, status):
        return status.getModificationTime()

    def is_directory_from_status(self, status) -> bool:
        return status.isDirectory()

    def get_readable_file_modification_time_from_status(self, status) -> datetime:
        return datetime.fromtimestamp(self.get_file_modification_time_from_status(status) / 1000.0)

    def is_directory_from_fs_path(self, fs_path) -> bool:
        status = self.get_file_status_from_fs_path(fs_path)
        return self.is_directory_from_status(status)

    def delete_file_from_fs_path(self, fs_path) -> bool:
        return self.easy_hadoop.fs.delete(fs_path, True)

    def file_exists_from_fs_path(self, fs_path) -> bool:
        return self.easy_hadoop.fs.exists(fs_path)

    def file_exists_from_path(self, path: str, is_relative: bool) -> bool:
        fs_path = self.get_fs_path(path, is_relative)
        return self.file_exists_from_fs_path(fs_path)

    def get_file_status_from_fs_path(self, fs_path):
        return self.easy_hadoop.fs.getFileStatus(fs_path)

    def get_fs_path(self, path: str, is_relative: bool):
        if is_relative:
            return self.get_fs_absolute_path(path)
        else:
            return self.easy_hadoop.make_path(path)

    def get_fs_path_from_file_path(self, file_path: EasySparkFilePath):
        return self.get_fs_path(file_path.path, file_path.is_relative)

    def get_fs_absolute_path(self, relative_path: str):
        fs_path = self.easy_hadoop.make_path(relative_path)
        return fs_path.makeQualified(self.easy_hadoop.fs.getUri(),
                                     self.easy_hadoop.fs.getWorkingDirectory())

    def list_statuses_from_fs_path(self, fs_path) -> list:
        return list(self.easy_hadoop.fs.listStatus(fs_path))

    def get_file_details_from_status(self, status) -> EasyFileDetails:
        is_directory = self.is_directory_from_status(status)
        file_name = self.get_file_name_from_status(status)
        fs_path = self.get_fs_path_from_status(status)
        full_path = self.get_file_full_path_from_status(status)

        if is_directory:
            return EasyFileDetails(fs_path, status, is_directory, file_name, full_path, None)

        readable_modified_time = self.get_readable_file_modification_time_from_status(status)

        return EasyFileDetails(fs_path, status, is_directory, file_name, full_path, readable_modified_time)

    def get_file_details_from_fs_path(self, fs_path) -> EasyFileDetails:
        status = self.get_file_status_from_fs_path(fs_path)
        return self.get_file_details_from_status(status)

    def get_file_details_from_path(self, path: str, is_relative: bool) -> EasyFileDetails:
        fs_path = self.get_fs_path(path, is_relative)
        return self.get_file_details_from_fs_path(fs_path)

    def list_path_details_from_fs_path(self, fs_path) -> Iterator[EasyFileDetails]:
        statuses = self.list_statuses_from_fs_path(fs_path)
        for status in statuses:
            yield self.get_file_details_from_status(status)

    def list_path_details_from_path(self, path: str, is_relative: bool) -> Iterator[EasyFileDetails]:
        fs_path = self.get_fs_path(path, is_relative)
        return self.list_path_details_from_fs_path(fs_path)



from typing import Generic, TypeVar, Type
import inspect
from easy_sql_data_access.easy_sql_data_access_factory import EasySQLDataAccessFactory
from easy_sql_data_access.easy_sql_data_entity_dict import EasySQLDataEntityDict
from easy_utils.easy_data_utils import EasyDataUtils
from easy_sql_data_access.easy_sql_data_access import EasySQLDataAccess

T = TypeVar('T')


class EasySQLDataEntityType(Generic[T]):
    def __init__(self, entity_type: Type[T], use_ctor: bool, data_access: EasySQLDataAccessFactory | EasySQLDataAccess,
                 table_name: str,
                 id_column_name: str = "Id", simple_map=True, add_lock=False):
        self.table_name = table_name
        self.id_column_name = id_column_name
        self.entity_type = entity_type
        self.class_parameters = inspect.signature(entity_type.__init__).parameters
        self.use_ctor = use_ctor
        self.simple_map = simple_map
        self.data_entity_dict = EasySQLDataEntityDict(data_access, table_name, id_column_name, add_lock)

    def update(self, id: any, entity: T):
        dict_record = EasyDataUtils.map_object_to_dict(entity)
        self.data_entity_dict.update(id, dict_record)

    def patch(self, id: any, entity: dict[str, any]):
        self.data_entity_dict.patch(id, entity)

    def insert(self, entity: T):
        dict_record = EasyDataUtils.map_object_to_dict(entity)
        self.data_entity_dict.insert(dict_record)

    def delete(self, id: any):
        self.data_entity_dict.delete(id)

    def get_list(self) -> list[T]:
        dict_list = self.data_entity_dict.get_list()
        return self.map_list_dict_to_object(dict_list)

    def get_list_with_filters(self, filters: dict[str, any]) -> list[T]:
        dict_list = self.data_entity_dict.get_list_with_filters(filters)
        return self.map_list_dict_to_object(dict_list)

    def get(self, id: any) -> T:
        dict_record = self.data_entity_dict.get(id)
        return self.map_dict_to_object(dict_record)

    def get_with_filters(self, filters: dict[str, any]) -> T:
        dict_record = self.data_entity_dict.get_with_filters(filters)
        return self.map_dict_to_object(dict_record)

    def map_dict_to_object(self, dict_record: dict[str, any]) -> T:
        if self.use_ctor:
            if self.simple_map:
                return EasyDataUtils.map_dict_to_object_ctor_simple(self.entity_type, dict_record)
            else:
                return EasyDataUtils.map_dict_to_object_from_params_ctor(self.entity_type, self.class_parameters,
                                                                         dict_record)
        else:
            return EasyDataUtils.map_dict_to_object(self.entity_type, dict_record)

    def map_list_dict_to_object(self, dict_list: list[dict[str, any]]) -> list[T]:
        if self.use_ctor:
            if self.simple_map:
                return EasyDataUtils.map_list_dict_to_object_ctor_simple(self.entity_type, dict_list)
            else:
                return EasyDataUtils.map_list_dict_to_object_from_params_ctor(self.entity_type, self.class_parameters,
                                                                              dict_list)
        else:
            return EasyDataUtils.map_list_dict_to_object(self.entity_type, dict_list)



from contextlib import contextmanager

import pyodbc


@contextmanager
def easy_open_data_connection(constr: str, autocommit: bool = True):
    con = None
    try:
        print('Connecting to SQL Server...')
        con = pyodbc.connect(constr, autocommit=autocommit)
        yield con
    finally:
        print('Closing connection...')
        if con:
            print('Closing connection Inside...')
            con.close()





from easy_spark.easy_spark_instance import EasySparkInstance
from easy_utils.easy_singleton import easy_singleton


@easy_singleton
class EasySparkCatalog:
    def __init__(self, spark):
        self.spark_instance = EasySparkInstance(spark)

    def table_exists_from_table_name(self, table_name: str, db_name: str) -> bool:
        return self.spark_instance.spark.catalog._jcatalog.tableExists(table_name, db_name)

    def table_exists_from_path(self, path: str) -> bool:
        return self.spark_instance.spark.catalog._jcatalog.tableExists(path)
